/*

Humboldt University Berlin
Department of Computer Science
Laboratory for Artificial Intelligence

AT Humboldt97
-------------
champion of RoboCup'97 - modified version for RoboCup'98

Copyright 1997,1998 Hans Dieter Burkhard, Markus Hannebauer, Jan Wendler, Kay Schroeter
                        (All rights reserved)

You may copy and distribute this program freely as long as you retain this notice.
If you make any changes or have any comments we would appreciate a message.

This programm was the winner-team of RoboCup97 (Simulation League).
It has been modified to cope with the new Soccer Server in RoboCup'98.

Program was developed by Hans Dieter Burkhard, Markus Hannebauer, Jan Wendler,
Kay Schroeter, Ralf Kuehnel, Pascal Mueller-Gugenberger, Adriana Foremniak,
Heike Mueller, Derrick Hepp and Amin Coja-Oghlan.

*/
///////////////////////////////////////////////////////////////////////////////
// advancedskills.c
///////////////////////////////////////////////////////////////////////////////
// Implementierung der erweiterten Faehigkeiten
///////////////////////////////////////////////////////////////////////////////

/* I included the following because I got problems with the pow-function of glibc 2.0.7 */
#define __NO_MATH_INLINES

#ifndef _ADVANCEDSKILLS_H_
#include "advancedskills.h"
#endif
#include "referee.h"
#include "Configuration.h"

AdvancedSkills::AdvancedSkills(Weltmodell* world, Sender* sender)
{
  myworld = world;
  mysender = sender;
}

AdvancedSkills::~AdvancedSkills()
{
}

/////////////////////////
// interne Operationen //
/////////////////////////

// Wissensbasis befragen
void AdvancedSkills::getinformation()
{
  Spieler_Selbst sim_ich_next;
  Ball sim_ball_next;
  
  myworld->berechne_relativ(ball_pos, sim_ich, sim_ball.frage_p());  
  max_power = (stamina < 100)? stamina : 100.0;
  if(max_power<50) max_power=50;

  v_ball=sim_ball.frage_v().norm();

  sim_ich_next=sim_ich;
  sim_ich_next.beschleunigen(max_power);
  sim_ich_next.simulieren(1);

  sim_ball_next=sim_ball;
  sim_ball_next.simulieren(1);

  myworld->berechne_relativ(ball_pos_next, sim_ich_next, sim_ball_next.frage_p());
}

// Informationen interpolieren
void AdvancedSkills::setinformation(int k, float a, float b)
{
  if (k == DASH){
    sim_ich.beschleunigen(a);
    stamina-=(a-20.0);
  }
  else if (k == KICK) {
    Vektor v_kick;
    v_kick=sim_ich.frage_richtung();
    v_kick.drehen(-deg_in_rad(b));
    sim_ball.beschleunigen(a, v_kick);
    stamina-=(a-20.0);
  }
  else if (k == TURN){
    ziel_winkel+=a;
    sim_ich.drehen(deg_in_rad(a));
    stamina+=20.0;
  }
  sim_ich.simulieren(1);
  sim_ich.sicher ();
  sim_ball.simulieren(1);
  sim_ball.sicher ();
  
}

////////////////////////////////////////////
// Einfache Faehigkeiten                  //
////////////////////////////////////////////

void AdvancedSkills::SimpleChangeView(int width, int quality, int stat=IGNORE){
  mysender->SetCommand(CHANGEVIEW,width,quality,stat);
}

void AdvancedSkills::SimpleMoveTo(double x, double y, int stat=IGNORE){
  mysender->SetCommand(MOVETO,x,-y,stat);
}

  int AdvancedSkills::SimpleTurn(double angle, int stat=IGNORE) {
    int command_number = 1;
    Spieler_Selbst player = myworld->frage_selbst ();
    Winkel real_angle = angle * (5.0 * player.frage_v ().norm () + 1.0);
    Winkel phi;
    Vektor neg_v;
    
#ifdef DEBUG
    cout << "SimpleTurn: START:" << endl;
    cout << "SimpleTurn: MyPos:     x: " << player.frage_p().x1 << " y: " << player.frage_p().x2 << endl;
    cout << "SimpleTurn: MySpeed:   x: " << player.frage_v().x1 << " y: " << player.frage_v().x2 << endl;
    cout << "SimpleTurn: MySpeed-Norm: " << player.frage_v ().norm () << endl;
    cout << "SimpleTurn: ViewAngle: " << rad_in_deg(player.frage_richtung().arg()) << endl;
    cout << "SimpleTurn: Wish_angle: " << angle << endl;
    cout << "SimpleTurn: Real_angle: " << real_angle << endl;
#endif
    if (fabs (real_angle) > 180.0) {
      phi = fabs (player.frage_v ().arg () - player.frage_richtung ().arg ());
      neg_v = cos (phi) * player.frage_v ();
#ifdef DEBUG
      cout << "SimpleTurn: RealAngle is to large" << endl;
      cout << "SimpleTurn: neg_v: x1 " << neg_v.x1 << " x2 " << neg_v.x2 << " Norm " << neg_v.norm () << endl;
      cout << "SimpleTurn: phi: " << rad_in_deg (phi) << endl;
#endif
      neg_v.setzen (neg_v.arg (), (neg_v.norm () > 0.3 ? 0.3 : neg_v.norm()));
#ifdef DEBUG
      cout << "SimpleTurn: neg_v nach max: x1 " << neg_v.x1 << " x2 " << neg_v.x2 << " Norm " << neg_v.norm () << endl;
#endif
      mysender->SetCommand (DASH, -neg_v.norm ()/DASH_POWER_RATE, IGNORE);
      command_number++;
      player.setze_v (player.frage_v () - neg_v);
      player.simulieren (1);
#ifdef DEBUG
      cout << "SimpleTurn: MyPos:     x: " << player.frage_p().x1 << " y: " << player.frage_p().x2 << endl;
      cout << "SimpleTurn: MySpeed:   x: " << player.frage_v().x1 << " y: " << player.frage_v().x2 << endl;
      cout << "SimpleTurn: MySpeed-Norm: " << player.frage_v ().norm () << endl;
#endif
      real_angle = angle * (5.0 * player.frage_v ().norm () + 1.0);
#ifdef DEBUG
      cout << "SimpleTurn: ViewAngle: " << rad_in_deg(player.frage_richtung().arg()) << endl;
      cout << "SimpleTurn: Real_angle (after dash): " << real_angle << endl;
#endif
    }
    mysender->SetCommand (TURN, -real_angle, stat);
    return command_number;
  }


  int AdvancedSkills::SimpleDashTurn(int power1, int power2, double angle, int stat=IGNORE) {
    int command_number = 0;

    Spieler_Selbst sim_ich = myworld->frage_selbst ();

    if(power1 != 0) {
      mysender->SetCommand(DASH, power1, stat);
      command_number++;
      sim_ich.beschleunigen(power1);
      sim_ich.simulieren(1);
    }
    
    if(power2 != 0) {
      mysender->SetCommand(DASH, power2, stat);
      command_number++;
      sim_ich.beschleunigen(power2);
      sim_ich.simulieren(1);
    }
    Winkel real_angle = angle * (5.0 * sim_ich.frage_v().norm() + 1.0);
    
#ifdef DEBUG
    cout << "SimpleDashTurn: START: abgesetzte dash`s: " << command_number << endl;
    cout << "SimpleDashTurn: MyPos:     x: " << sim_ich.frage_p().x1 << " y: " << sim_ich.frage_p().x2 << endl;
    cout << "SimpleDashTurn: MySpeed:   x: " << sim_ich.frage_v().x1 << " y: " << sim_ich.frage_v().x2 << endl;
    cout << "SimpleDashTurn: MySpeed-Norm: " << sim_ich.frage_v ().norm () << endl;
    cout << "SimpleDashTurn: ViewAngle: " << rad_in_deg(sim_ich.frage_richtung().arg()) << endl;
    cout << "SimpleDashTurn: Wish_angle: " << angle << endl;
    cout << "SimpleDashTurn: Real_angle: " << real_angle << endl;
#endif
    while(fabs (real_angle) > 180.0) {

      mysender->SetCommand(TURN, -real_angle, stat);
      angle -= 180.0 / (5.0 * sim_ich.frage_v().norm() + 1.0);
      command_number++;
      sim_ich.drehen(deg_in_rad(angle));
      sim_ich.simulieren(1);

      real_angle = angle * (5.0 * sim_ich.frage_v().norm() + 1.0);
#ifdef DEBUG
      cout << "SimpleDashTurn: RealAngle is to large, I will split him (i turn the first 180.0 degree)" << endl;
      cout << "SimpleDashTurn: ViewAngle: " << rad_in_deg(sim_ich.frage_richtung().arg()) << endl;
      cout << "SimpleDashTurn: Wish_angle: (which part is left) " << angle << endl;
      cout << "SimpleDashTurn: Real_angle: (which part is left) " << real_angle << endl;
#endif
    }

    mysender->SetCommand(TURN, -real_angle, stat);
    command_number++;
    sim_ich.drehen(deg_in_rad(angle));
    sim_ich.simulieren(1);

#ifdef DEBUG
    cout << "SimpleDashTurn: Ready:" << endl;
    cout << "SimpleDashTurn: MyPos:     x: " << sim_ich.frage_p().x1 << " y: " << sim_ich.frage_p().x2 << endl;
    cout << "SimpleDashTurn: MySpeed:   x: " << sim_ich.frage_v().x1 << " y: " << sim_ich.frage_v().x2 << endl;
    cout << "SimpleDashTurn: ViewAngle: " << rad_in_deg(sim_ich.frage_richtung().arg()) << endl;
#endif

    return command_number;
  }


void AdvancedSkills::SimpleDash(double power, int stat=IGNORE){
    mysender->SetCommand(DASH,power,stat);
}

void AdvancedSkills::SimpleKick(double power, double angle, int stat=IGNORE){
  mysender->SetCommand(KICK,power,-angle,stat);
}

void AdvancedSkills::SimpleSay(char *message, int stat=IGNORE){
  mysender->SetCommand(SAY,message,stat);
}
  
int sort_pos(Relative_Koordinate *a,Relative_Koordinate *b)
{
  if(fabs(a->richtung)<=fabs(b->richtung)) return -1;
  else return 1;
}

int AdvancedSkills::ausweichen(Winkel &gamma){ // Winkel in Grad
  Relative_Koordinate spieler_pos[20],delta;
  int anzahl=0,i,first_pos_index,first_neg_index;
  Spieler_Selbst local_sim_ich=myworld->frage_selbst();
  Winkel diff_l,diff_r;

  local_sim_ich.drehen(deg_in_rad(gamma));
  local_sim_ich.simulieren(1);

  for(i=0;i<11;i++){
    if(myworld->frage_spieler(i,team_sie).ist_unsicher()) continue; // nicht gesehen
    myworld->berechne_relativ(delta,local_sim_ich,myworld->frage_spieler(i,team_sie).frage_p());
    if(delta.d>10) continue; // zu weit weg
    spieler_pos[anzahl]=delta;
    anzahl++;
  }
  for(i=0;i<10;i++){
    if(myworld->frage_spieler(i,team_wir).ist_unsicher()) continue; // nicht gesehen
    myworld->berechne_relativ(delta,local_sim_ich,myworld->frage_spieler(i,team_wir).frage_p());
    if(delta.d>10) continue; // zu weit weg
    spieler_pos[anzahl]=delta;
    anzahl++;
  }

  if(anzahl==0) return 0; // alles frei
  
  qsort(spieler_pos,anzahl,sizeof(Relative_Koordinate),sort_pos);
  for(first_pos_index=0;first_pos_index<anzahl;first_pos_index++){
    if(spieler_pos[first_pos_index].richtung>=0) break;
  }
  for(first_neg_index=0;first_neg_index<anzahl;first_neg_index++){
    if(spieler_pos[first_neg_index].richtung<0) break;
  }
  
  if(spieler_pos[first_neg_index].d<1.7) spieler_pos[first_neg_index].d=1.7; // sollte eigentlich nicht vorkommen, aber sicher ist sicher
  if(spieler_pos[first_pos_index].d<1.7) spieler_pos[first_pos_index].d=1.7; 

  if(first_pos_index==anzahl){ // rechts steht nichts -> falls noetig nach rechts ausweichen
    diff_l=rad_in_deg(asin(1.7/spieler_pos[first_neg_index].d)+spieler_pos[first_neg_index].richtung);
    if(diff_l<=0.0) return 0; // genug Platz
    else{
      gamma+=diff_l;
      return 1;
    }
  }
  if(first_neg_index==anzahl){ // links steht nichts -> falls noetig nach links ausweichen 
    diff_r=rad_in_deg(spieler_pos[first_pos_index].richtung-asin(1.7/spieler_pos[first_pos_index].d));
    if(diff_r>=0.0) return 0; // genug Platz
    else{
      gamma+=diff_r;
      return 1;
    }
  }
   diff_l=rad_in_deg(asin(1.7/spieler_pos[first_neg_index].d)+spieler_pos[first_neg_index].richtung);
  diff_r=rad_in_deg(spieler_pos[first_pos_index].richtung-asin(1.7/spieler_pos[first_pos_index].d));
  // In der Mitte durchlaufen;
  if((diff_l<=0.0) && (diff_r>=0.0)){ // rechts und links genug Platz -> Durchlaufen
    return 0;
  }
  else if (diff_r>=0.0 && diff_r-diff_l>=0.0){ // rechts ist genug Platz zum ausweichen
   gamma+=diff_l;				     
   return 1;
  }
  else if (diff_l<=0.0 && diff_l+diff_r<=0.0){ // links ist genug Platz zum ausweichen
    gamma+=diff_r;
    return 1;
  }

  // Luecke mit der geringsten Abweichung von Zielrichtung suchen;  
  for(i=2;i<anzahl;i++){
    if(i==first_pos_index) continue;
    if(i==first_neg_index) continue;
    if(spieler_pos[i].d<1.7) spieler_pos[i].d=1.7;

    if(spieler_pos[i].richtung<0){  
      // links;
      diff_l=rad_in_deg(spieler_pos[i].richtung+asin(1.7/spieler_pos[i].d));
      diff_r=rad_in_deg(spieler_pos[first_neg_index].richtung-asin(1.7/spieler_pos[first_neg_index].d));
      if(diff_l>=0 || diff_r<diff_l){
        // zu wenig Platz
	first_neg_index=i;
	continue;
      }
      gamma+=(diff_r+(diff_l-diff_r)/2);
      return 1;      
    }
    else{
      // rechts;      
      diff_l=rad_in_deg(spieler_pos[first_pos_index].richtung+asin(1.7/spieler_pos[first_pos_index].d));
      diff_r=rad_in_deg(spieler_pos[i].richtung-asin(1.7/spieler_pos[i].d));
      if(diff_r<=0 || diff_r<diff_l){
        // zu wenig Platz;
	first_pos_index=i;
	continue;
      }
      gamma+=(diff_l+(diff_r-diff_l)/2);
      return 1;
    }
  }

  // versuche am letzten den ich gesehen habe vorbei zu laufen;
  diff_l=rad_in_deg(spieler_pos[first_neg_index].richtung-asin(1.7/spieler_pos[first_neg_index].d));
  diff_r=rad_in_deg(spieler_pos[first_pos_index].richtung+asin(1.7/spieler_pos[first_pos_index].d));
  if(diff_l>-180){
    gamma+=diff_l;
    return 1;
  }
  if(diff_r<180){
    gamma+=diff_r;
    return 1;
  }
  return 0;
}

////////////////////////////////////////////
// Erweiterte Faehigkeiten: Ballkontrolle //
////////////////////////////////////////////

void AdvancedSkills::Dribbeln(Winkel zielrichtung)
{
  sim_ich=myworld->frage_selbst();
  sim_ball=myworld->frage_ball();
  ziel_winkel=zielrichtung;
  stamina=myworld->frage_selbst().frage_stamina();
  int status=IGNORE;

  mysender->InitializeQueue();
  // Markus-Aenderung damit Turn-Kommandos eingerechnet werden
int durchlauf = 0;
  while (durchlauf <= 2)
  {
    if(durchlauf==2) status=LAST;

    if(stamina>600) staerke=1.0;
    else if(stamina>200) staerke=0.6;
    else staerke=0.2;
    
    getinformation();
    
    /*if((ball_pos.d<=1.8) && (fabs(rad_in_deg(ball_pos.richtung))>90 || v_ball>2.0*staerke)){
      // Ball liegt falsch oder ist sehr schnell
      cerr << "Ballpos.d: " << ball_pos.d << " ball_pos.richtung: " << rad_in_deg(ball_pos.richtung) << " vBall " << v_ball << " Zielwinkel: " << ziel_winkel << "\n";
      Kick(-ziel_winkel,2.0);
      return;
    }

    else*/
    if((ball_pos.d<=1.8) && (fabs(rad_in_deg(ball_pos_next.richtung))>90 || v_ball<0.1)){
      // kann diesmal kicken, aber beim naechsten mal nicht mehr bzw. Ball bewegt sich nicht
      mysender->SetCommand(KICK,max_power*staerke*0.8,ziel_winkel,status);
      setinformation(KICK,max_power*staerke*0.8,ziel_winkel);
      durchlauf++;
    }
    else
    {
      if(fabs(rad_in_deg(ball_pos.richtung))>5.0)
      {
        //bin nicht im Kickbereich und Richtung stimmt nicht    
        //mysender->SetCommand(TURN,rad_in_deg(ball_pos.richtung),status);
        durchlauf += SimpleTurn (-rad_in_deg (ball_pos.richtung), status);
        setinformation(TURN,rad_in_deg(ball_pos.richtung));
      }
      else
      {
        mysender->SetCommand(DASH, max_power, status);
        setinformation(DASH,max_power);
        durchlauf++;
      }
    }
  }
}

void AdvancedSkills::GotoBall(){
  int schritte=0,count,min;
  double turn=0.0,last_dash=0.0, stamina_used=0.0;
  Ball help_ball=myworld->frage_ball(); // brauch ich wegen irgendeinem const !!?

  mysender->InitializeQueue();				     
  if(help_ball.frage_wert()>3){       // Habe den Ball schon lange nicht mehr gesehen ! 
    //mysender->SetCommand(TURN,88.0); // Wo laufe ich eigentlich hin ???
    SimpleTurn (-88.0, IGNORE);
    return;
  }
  
  schritte=BestWay2Ball(myworld->frage_ball(),myworld->frage_selbst(),&turn,&last_dash,&stamina_used);
  min=((schritte<3)? schritte: 3);
  for(count=1; count<=min; count++){
    if(count==1 && fabs(turn)>=1.0){
      //mysender->SetCommand(TURN,(float)turn);
      SimpleTurn ((float) (-turn), IGNORE);
    }
    else if(count==schritte){
      if(last_dash!=0) mysender->SetCommand(DASH,(float)last_dash);
    }
    else{
      mysender->SetCommand(DASH,(float)100.0);
    }
  }
}

void AdvancedSkills::GotoInterceptPosition (Punkt ziel){
  Relative_Koordinate delta;
  Winkel gamma;
  Spieler_Selbst sim_ich;
  int count=0,ausgewichen=0,dash_next=0;
  
  sim_ich=myworld->frage_selbst();
  
#ifdef DEBUG
  cout << "GotoInterceptPosition: START" << endl;
#endif

  // Markus-Aenderung, damit Anzahl der Turn-Kommandos eingerechnet wird;
  while (count <= 2) {
    myworld->berechne_relativ(delta,sim_ich,ziel);
    if(delta.d<0.5) return; // von Markus geaendert;
    
    gamma=rad_in_deg(delta.richtung);
    ausgewichen=ausweichen(gamma); // falls ausgewichen nach moeglichkeit 2x dashen;
    if(dash_next==0 && (fabs(gamma)>10.0 || ausgewichen==1)) {
      //mysender->SetCommand(TURN,(float)gamma);
      count += SimpleTurn ((float)(-gamma));
      sim_ich.drehen(delta.richtung);
      if (ausgewichen==1) dash_next=2;
    }
    else if(delta.d>=1.0) {
      mysender->SetCommand(DASH,(float)100.0);	
      sim_ich.beschleunigen(100.0);
      if(dash_next>0) dash_next--;
      count++;
    }
    else {
      mysender->SetCommand(DASH, (float)delta.d*10, IGNORE); // das wars - letztes dash;
      return;
    } 
    sim_ich.simulieren(1); 
    #ifdef DEBUG
    cout << "GotoPosition: simulated: " << count << " Steps" << endl;
    cout << "GotoPosition: MyPos: x: " << sim_ich.frage_p().x1 << " y: " << sim_ich.frage_p().x2 << endl;
    cout << "GotoPosition: MySpeed: x: " << sim_ich.frage_v().x1 << " y: " << sim_ich.frage_v().x2 << endl;
    cout << "GotoPosition: DestPos: x: " << ziel.x1 << " y: " << ziel.x2 << endl;
    #endif
  }  
}

void AdvancedSkills::GotoPosition (Punkt ziel, int stat=IGNORE) {
  Relative_Koordinate delta, delta2;
  Winkel gamma;
  Spieler_Selbst sim_ich, gegner;
  int count;
  double abseitsX;
  
  sim_ich=myworld->frage_selbst();

#ifdef DEBUG
  cout << "GotoPosition: START" << endl;
#endif
  
  // Markus-Aenderung, damit Anzahl der Turn-Kommandos eingerechnet wird;
  count = 0;
  if (myworld->blockierenderGegnerIndex>=0 && myworld->blockierenderGegnerIndex<11){
    gegner=(Spieler_Selbst&) myworld->frage_spieler(myworld->blockierenderGegnerIndex, team_sie);
    myworld->berechne_relativ(delta,sim_ich,ziel);
    myworld->berechne_relativ(delta2,sim_ich,gegner.frage_p());
    if (delta2.d<2.0 && rad_in_deg(fabs(delta.richtung-delta2.richtung)<85.0)){
      SimpleTurn((float) (signum(delta2.richtung))*90);
      SimpleDash((float) 100);
      SimpleDash((float) 100);
    }
  return;
  }
  
  // Ziel liegt im Abseits und;
  // Ball hat die Abseitslinie noch nicht passiert;
  
  while (count <= 2) {
    myworld->berechne_relativ(delta,sim_ich,ziel);
//    if(delta.d < 0.5) return; 
    if(delta.d < 1.0) return; 
		
    gamma=rad_in_deg(delta.richtung);    
    if(fabs(gamma)>10.0) {
      //mysender->SetCommand(TURN,(float)gamma);
      count += SimpleTurn ((float)(-gamma));
      sim_ich.drehen(delta.richtung);
    }
    else if(delta.d>=1.0) { 
      mysender->SetCommand(DASH,(float)100.0);
      sim_ich.beschleunigen(100.0);
      count++;
    }
    else {
      mysender->SetCommand(DASH, (float)delta.d*10, stat);
      count++;
      return;
    } 
    sim_ich.simulieren(1); 
    #ifdef DEBUG
    cout << "GotoPosition: simulated: " << count << " Steps" << endl;
    cout << "GotoPosition: MyPos: x: " << sim_ich.frage_p().x1 << " y: " << sim_ich.frage_p().x2 << endl;
    cout << "GotoPosition: MySpeed: x: " << sim_ich.frage_v().x1 << " y: " << sim_ich.frage_v().x2 << endl;
    cout << "GotoPosition: DestPos: x: " << ziel.x1 << " y: " << ziel.x2 << endl;
    #endif
  }  
}

int AdvancedSkills::BestWay2Ball(const Ball ziel, const Spieler_Selbst ich, double *turn, double *last_dash, double *stamina_used){
  int schritte=0;
  *turn=0.0;
  *last_dash=0.0;
  Ball sim_ball=ziel;
  Spieler_Selbst sim_ich,sim_ich_kopie=ich;

  Relative_Koordinate delta;
  Punkt p;
  Winkel gamma;
 
  while(1){
    /* Anzahl der Schritte abschaetzen - Bewegung mit max. Geschwindigkeit */
    
    if ( pow(sim_ball.frage_p().x1-sim_ich.frage_p().x1,2) + pow(sim_ball.frage_p().x2-sim_ich.frage_p().x2,2) >= pow(schritte * globalConf->playerSpeedMax,2)){
      break;
    }
    sim_ball.simulieren(1);
    sim_ball.sicher ();
    if(schritte>100) return -1;
    schritte++;
  }
  
  while(1){
    sim_ich=sim_ich_kopie;
    *stamina_used=0;
    for(int i=0; i<schritte; i++){
      
      myworld->berechne_relativ(delta,sim_ich,sim_ball.frage_p());
      gamma=rad_in_deg(delta.richtung);
      
      if(i==0 && (gamma<-0.1 || gamma>0.1)){ /* 1. Kommando */
	sim_ich.drehen(delta.richtung);
	*turn=gamma;
	*stamina_used=(*stamina_used)-20;
      }
      else{
	if(delta.d>=1.0){
	  *last_dash=100.0;
	  sim_ich.beschleunigen(100.0);
	  *stamina_used=(*stamina_used)-20+100;
	}
	else{
	  *last_dash=delta.d*100;
	  sim_ich.beschleunigen(delta.d*100);
	  *stamina_used=(*stamina_used)-20+delta.d*100;
	}
      }
      sim_ich.simulieren(1);
      sim_ich.sicher ();
    }
    myworld->berechne_relativ(delta,sim_ich,sim_ball.frage_p());
    if(delta.d<=1.8) return schritte;
    sim_ball.simulieren(1);
    sim_ball.sicher ();
    if(schritte>100) return -1;
    schritte++;
  }
}

// Jan

/**
 * Universelle Kick-Routine. Die Parameter sind der Winkel in dem der Spieler
 * den Ball treten soll (relativ von seiner Blickrichtung aus) und die
 * Kraft mit der er den Ball effektiv nach Ausfuehrung der Kick-Routine
 * getreten hat. Der power-Wert kann dabei ohne weiteres ueber 100 gewaehlt
 * werden. Da dieses Kick meist mehrere atomare kick's ausfuehrt werden
 * power-Werte bis zu 200 auch im allgemeinen erreicht. Falls es nicht
 * moeglich ist den angegebenen power-Wert zu erreichen, dann wird dieser
 * maximiert. Bei gut liegenden Ball ist es auch moeglich einen power-Wert
 * von ueber 300 zu erreichen. Negative power-Werte sind ueberfluessig
 * und werden so behandelt als ob ein Kick mit dem entsprechenden positiven
 * power-Wert in die entgegengesetzte Richtung erfolgte.
 * In einigen F"allen (wenn der Ball zu schnell ist) kann es vorkommen das
 * der eigene Spieler den Ball nicht mehr kontrollieren kann. Dies bedeutet
 * das es dem Spieler nicht gelingt den Ball in die mit degree angegebene 
 * Richtung zu lenken. In solchen F"allen wird der Ball abgelenkt.
 */
  void AdvancedSkills::Kick(const Winkel degree, const Koerper power, int state=LAST, int maxKicks=3) {

      sim_ball = myworld->frage_ball();
      sim_ich = myworld->frage_selbst();

      #ifdef DEBUG
      cout << "Jan: TargetSpeedRelativ: Angle: " << degree << " Power: " << power << endl;
      cout << "Jan: Blickwinkel: " << rad_in_deg(sim_ich.frage_richtung().arg()) << endl;
      #endif

      Geschwindigkeit targetSpeed;
      targetSpeed.setzen(deg_in_rad(degree)+sim_ich.frage_richtung().arg(), power*KICK_POWER_RATE);    

      #ifdef DEBUG
      cout << "Jan: TargetSpeedAbsolut: Angle: " << rad_in_deg(targetSpeed.arg()) << " Power: " << targetSpeed.norm()/KICK_POWER_RATE << endl;
      #endif

      KickInDir(targetSpeed, state, maxKicks);
  }

/**
 * This is the ultimate do-it-all skill. It kicks the ball in a
 * wanted direction with a wanted speed. (If it cannot reach the desired
 * speed, it tries to maximize it.
 * It takes as many ticks as it needs. It can
 * be entered and aborted at any point.
 * Input: target direction. Length of vector is interpreted as desired speed.
 *        Note that the desired speed is not necessarily reached
 */
  void AdvancedSkills::KickInDir(const Geschwindigkeit targetSpeed, int state=LAST, int maxKicks=3) {
      // First see if 1 kick can give ball desired direction;
      Vektor kickDir;
      Ball projectedBall;
      int kickNumber = 0;
      
      sim_ball = myworld->frage_ball();
      sim_ich = myworld->frage_selbst();
      
//      mysender->InitializeQueue();

#ifdef DEBUG
      cout << endl << endl;
      cout << "KickInDir: START: Time         : " << sim_ich.frage_jetzt() << endl;
      cout << "KickInDir: START: myPos    : x1: " << sim_ich.frage_p().x1 << " x2: " << sim_ich.frage_p().x2 << endl;
      cout << "KickInDir: START: mySpeed  : x1: " << sim_ich.frage_v().x1 << " x2: " << sim_ich.frage_v().x2 << endl;
      cout << "KickInDir: START: ballPos  : x1: " << sim_ball.frage_p().x1 << " x2: " << sim_ball.frage_p().x2 << endl;
      cout << "KickInDir: START: ballSpeed: x1: " << sim_ball.frage_v().x1 << " x2: " << sim_ball.frage_v().x2 << endl;
      cout << "KickInDir: TargetSpeed: x1: " << targetSpeed.x1 << " x2: " << targetSpeed.x2 << endl;
#endif

      kickDir = computeKickDir(targetSpeed);
      
      
      if(kickDir == NULL_VEKTOR) {
        // Diese Behandlung muss noch anders erfolgen, Analyse lieber Ball blocken oder laufen lassen;
        // Zur Zeit wird ein Ablenken des Balls unternommen;
        // vielleicht ist es gut sich anzuschiessen, falls das geht und dann den Ball zu spielen;
	
#ifdef DEBUG
        cout << "KickInDir: NULL_VEKTOR calculated" << endl;
#endif
        // no chance to kick ball in desired direction;
        //	kickDir.setzen(targetSpeed.arg() + PI/2, MAX_BALL_ACC);  // Normalenvektor zum Zielvektor bestimmen;
        kickDir.setzen(sim_ball.frage_v().arg() + PI/2, MAX_BALL_ACC);  // Normalenvektor zum BallSpeedvektor bestimmen;
	
        setKick(kickDir);
	
        // now check ball position again and see if we can kick again to;
	// obtain desired speed;
	if(distanceBetween(sim_ball, (Spieler)sim_ich) < KICK_RANGE) {
	  kick(kickDir);
	  if(++kickNumber == maxKicks) return;
	  
	  kickDir = computeKickDir(targetSpeed);
	  if(kickDir == NULL_VEKTOR) {
	    // wieder eine entsprechende Ausnahmebehandlung (noch aendern);
	    
            kickDir.setzen(sim_ball.frage_v().arg() + PI/2, MAX_BALL_ACC);  // Normalenvektor zum BallSpeedvektor bestimmen;
	  }
	  setKick(kickDir);
	  kick(kickDir, state);
	  if(++kickNumber == maxKicks) return;
	}
	else {
	  kick(kickDir, state);
	  if(++kickNumber == maxKicks) return;
	}
      }
      
      
      
      else {  
        // kickDir != NULL_VEKTOR;
        // it is possible to kick ball in desired direction;
        // but first check if player is in the way;
        // we have to use a copy of ball for the test;
        // because we need the original speed to recompute kickDir;

        projectedBall = sim_ball;
        projectedBall.setze_v(projectedBall.frage_v() + kickDir);

        if( collides(projectedBall, (Spieler)sim_ich) >= 0 ) {
          // if yes, add intermediate kick;
#ifdef DEBUG
          cout << "KickInDir: calculate intermediate target" << endl;
#endif

	  Vektor kickDirCopy = kickDir;
	  Vektor firstTarget = computeIntermediateTarget(targetSpeed);
	  kickDir = computeKickDir(firstTarget);
	  if(kickDir == NULL_VEKTOR) {
	    // ich kann kein richtiges Zwischenziel bestimmen -> soll er ruhig;
	    // auf mich treffen;
	    kickDir = kickDirCopy;
	    setKick(kickDir);
	    kick(kickDir);
	    if(++kickNumber == maxKicks) return;
	  }
	  else {
	    testAssertion(kickDir);

	    // Ball auf Zwischenposition schiessen;
	    setKick(kickDir);
	    kick(kickDir);
	    if(++kickNumber == maxKicks) return;
	  }
	  
	  // recompute kickDir
	  kickDir = computeKickDir(targetSpeed);
	  if(kickDir == NULL_VEKTOR) {
	    return;
	  }
	  else testAssertion(kickDir);
	}

	// Ball schiessen;
	setKick(kickDir);

	// check ball position again and see if we can kick again to
	// obtain desired speed
	if( sim_ball.frage_v().norm() < targetSpeed.norm() &&
	          distanceBetween(sim_ball, (Spieler)sim_ich) < KICK_RANGE ) {
      #ifdef DEBUG
      cout << "kickinDir: another kick comes" << endl;
      #endif

	  kick(kickDir);
	  if(++kickNumber == maxKicks) return;

	  kickDir = computeKickDir(targetSpeed);
	  if(kickDir == NULL_VEKTOR) {
	    return;
	  }
	  else testAssertion(kickDir);

	  setKick(kickDir);
	  kick(kickDir, state);
	  if(++kickNumber == maxKicks) return;
	}
	else {
	  kick(kickDir, state);
	  if(++kickNumber == maxKicks) return;
	}
      }
      #ifdef DEBUG
      cout << "KickInDir: ENDE: myPos    : x1: " << sim_ich.frage_p().x1 << " x2: " << sim_ich.frage_p().x2 << endl;
      cout << "KickInDir: ENDE: mySpeed  : x1: " << sim_ich.frage_v().x1 << " x2: " << sim_ich.frage_v().x2 << endl;
      cout << "KickInDir: ENDE: ballPos  : x1: " << sim_ball.frage_p().x1 << " x2: " << sim_ball.frage_p().x2 << endl;
      cout << "KickInDir: ENDE: ballSpeed: x1: " << sim_ball.frage_v().x1 << " x2: " << sim_ball.frage_v().x2 << endl;
      cout << "KickInDir: ENDE: ballSpeed: power: " << sim_ball.frage_v().norm() << " angle: " << rad_in_deg(sim_ball.frage_v().arg()) << endl;
      #endif
      
      return;
  }


/**
 * computeKickDir computes a kickVector so that, if the ball is kicked
 * according to this vector, afterwards the ball has the speed vector
 * targetSpeed.
 * That is, after 1 tick the ball is at currentBallPos+targetSpeed
 * According to this vector means: if kickDir is the resulting vector
 * kick( kickDir.getLength(), kickDir.getAngle() )
 * ( the necessary power is encoded in the length of the vector )
 * If only the direction of the targetSpeed can be produced,
 * computeKickDir still returns a vector.
 * If not even the desired direction can be produced, null is returned.
 */
  Vektor AdvancedSkills::computeKickDir(const Geschwindigkeit targetSpeed) {
      Geschwindigkeit k, v, z;
      z = targetSpeed;
      v = sim_ball.frage_v();
      k = z-v;  // k = z.relativTo(v);
      
      #ifdef DEBUG
      cout << "computeKickDir: TargetSpeed: x1: " << z.x1 << " x2: " << z.x2 << endl;
      cout << "computeKickDir: BallSpeed  : x1: " << v.x1 << " x2: " << v.x2 << endl;
      cout << "computeKickDir: KickSpeed  : x1: " << k.x1 << " x2: " << k.x2 << endl;
      cout << endl;
      #endif
      if(k.norm() > MAX_BALL_ACC) {
	// adapt length of z (z := z*p) , p is a double

	double p1, p2, p;
	double discriminant;
	double a = z*z;  // Skalarmultiplikation;

	// special case a == 0 (no quadratic equation!);
	if (a <  FLOAT_ERROR) { 
	  k = -v;
	  if(k.norm() > MAX_BALL_ACC + FLOAT_ERROR)
	    k.setzen(k.arg(), MAX_BALL_ACC);
	}
	else {
	  // now solve quadratic equation;
	  double b = 2 * (z*v);
	  double c = (v*v) - (MAX_BALL_ACC * MAX_BALL_ACC);
	  discriminant = b*b - 4*a*c;

	  if(discriminant < 0)  // target direction cannot be produced;
	    return NULL_VEKTOR;

	  discriminant = sqrt(discriminant);
	  a *= 2;
	  p1 = (b + discriminant) / a;
	  p2 = (b - discriminant) / a;
	  p = (p1 > p2) ? p1 : p2;

	  z.setzen(z.arg(), z.norm()*p);

	  k = z-v;
	}
  #ifdef DEBUG
  cout << "computeKickDir: KickSpeed was to long" << endl;
  cout << "computeKickDir: KickSpeedNew: x1: " << k.x1 << " x2: " << k.x2 << endl;
  #endif  

      }

      //assert(k.norm() <= MAX_BALL_ACC + FLOAT_ERROR);
#ifdef DEBUG
      cout << "computeKickDir: KickSpeedNorm: " << k.norm() << endl;
#endif
      
      return k;
      
  }


/**
 * computes an intermediate target that directly to the right of the
 * player (with respect to the direction of targetSpeed).
 * Use computeKickDir with the resulting vector to get the kickVector
 * to the intermediate target.
 * This function always returns a vector (provided that 
 * computeKickDir(targetSpeed) does not return null).
 */
  Vektor AdvancedSkills::computeIntermediateTarget(const Geschwindigkeit targetSpeed) {
      Geschwindigkeit result;
      Punkt myPosRelativeToBallPos;
      
      myPosRelativeToBallPos = sim_ich.frage_p()+sim_ich.frage_v();  // meine Pos. im naechsten Takt;
      myPosRelativeToBallPos -= sim_ball.frage_p();  // meine Pos. relativ zum Ball;

      #ifdef DEBUG
      cout << "computeIntermediateTarget: TargetSpeed: x1: " << targetSpeed.x1 << " x2: " << targetSpeed.x2 << endl;
      #endif

      Winkel angle = myPosRelativeToBallPos.arg() - targetSpeed.arg();
      if(angle > PI) angle -= 2*PI;
      if(angle < -PI) angle += 2*PI;
      #ifdef DEBUG
      cout << "computeIntermediateTarget: angle(myPos+mySpeed-ballPos)-angle(targetSpeed): angle: " << rad_in_deg(angle) << endl;
      #endif
      if(angle >= 0 && angle <= PI/2)  // (0,90)
        result.setzen(targetSpeed.arg() - PI/2, KICK_RANGE);  // Normalenvektor zum Zielvektor bestimmen;
      else if(angle >= -PI/2 && angle <= 0)  // (-90,0)
        result.setzen(targetSpeed.arg() + PI/2, KICK_RANGE);  // Normalenvektor zum Zielvektor bestimmen;
      else {
	result.x1 = 0;
	result.x2 = 0;
        #ifdef DEBUG
        cout << "WARNUNG: Ich spiele mich selber an" << endl << endl;
        #endif
        ;
      }
      #ifdef DEBUG
      cout << "computeIntermediateTarget: result(normalVektor) : x1: " << result.x1 << " x2: " << result.x2 << endl;
      cout << "computeIntermediateTarget: myPos+mySpeed-ballPos: x1: " << myPosRelativeToBallPos.x1 << " x2: " << myPosRelativeToBallPos.x2 << endl;
      #endif
  
      result += myPosRelativeToBallPos;  // der resultierende Vektor;

      result.setzen(result.arg(), result.norm() * 0.7);  // ein wenig verkuerzt;
      
      #ifdef DEBUG
      cout << "computeIntermediateTarget: result (intermediateTarget): x1: " << result.x1 << " x2: " << result.x2 << endl;
      #endif

      return result;
  }

/**
 * calculate the distance between the player and the ball
 */
  Koerper AdvancedSkills::distanceBetween(Ball ball, Spieler player) {
      return (player.frage_p() - ball.frage_p()).norm();
  }

/**
 * Gibt die Zahl der Ticks zurueck nach denen die Objekte kollidieren
 * (also evtl. auch 0), oder -1, wenn sie nicht kollidieren.
 * Dabei wird der SpeedDecay und der Radius der Objekte mit 
 * in Betracht gezogen.
 */
  int AdvancedSkills::collides(Ball ball, Spieler player) {
      Koerper dNew, dOld;
      int t=0;

      // Distanz jetzt;
      dNew = distanceBetween(ball, player);
      dOld = dNew + 1;    // dOld muss nur groesser sein;

      // jetzt simulieren;
      while(dNew > PLAYER_RADIUS + BALL_RADIUS && dNew < dOld) {
	player.simulieren(1);
	player.sicher ();
	ball.simulieren(1);
	ball.sicher ();	

	dOld = dNew;
	dNew = distanceBetween(ball, player);
	t++;
      }
      if(dNew <= PLAYER_RADIUS + BALL_RADIUS)
	return t;
      else 
	return -1;
  }

/**
 * kapselt das kick-Kommando
 */
  void AdvancedSkills::kick(Geschwindigkeit kickDir, int stat=IGNORE) {

      double power = kickDir.norm();
      double angle = kickDir.arg();

      // Werte fuers kick-Kommando anpassen;
      power = power/KICK_POWER_RATE;
      angle = rad_in_deg(angle)-rad_in_deg(sim_ich.frage_richtung().arg());

      #ifdef DEBUG
      cout << "kick: kickSpeed: Power: " << power << " Angle: " << angle <<  "   (Blickwinkel: " << rad_in_deg(sim_ich.frage_richtung().arg()) << ")" << endl << endl;
      #endif

      // nutze jetzt das SimpleKick Kommando;
      // ins System von Noda umwandeln;
      // angle = -angle;
      // mysender->SetCommand(KICK, power, angle, stat);

      SimpleKick(power, angle, stat);

      #ifdef DEBUG
      cout << "kick: MIDDLE: Time         : " << sim_ich.frage_jetzt() << endl;
      cout << "kick: MIDDLE: myPos    : x1: " << sim_ich.frage_p().x1 << " x2: " << sim_ich.frage_p().x2 << endl;
      cout << "kick: MIDDLE: mySpeed  : x1: " << sim_ich.frage_v().x1 << " x2: " << sim_ich.frage_v().x2 << endl;
      cout << "kick: MIDDLE: ballPos  : x1: " << sim_ball.frage_p().x1 << " x2: " << sim_ball.frage_p().x2 << endl;
      cout << "kick: MIDDLE: ballSpeed: x1: " << sim_ball.frage_v().x1 << " x2: " << sim_ball.frage_v().x2 << endl;
      #endif
  }

/**
 * kapselt das simulieren des kick-Kommandos
 */
  void AdvancedSkills::setKick(Geschwindigkeit kickDir) {

//      cout << "setKick: BEFORE:" << endl;
//      cout << "setKick: myPos    : x1: " << sim_ich.frage_p().x1 << " x2: " << sim_ich.frage_p().x2 << endl;
//      cout << "setKick: mySpeed  : x1: " << sim_ich.frage_v().x1 << " x2: " << sim_ich.frage_v().x2 << endl;
//      cout << "setKick: ballPos  : x1: " << sim_ball.frage_p().x1 << " x2: " << sim_ball.frage_p().x2 << endl;
//      cout << "setKick: ballSpeed: x1: " << sim_ball.frage_v().x1 << " x2: " << sim_ball.frage_v().x2 << endl;
    #ifdef DEBUG 
    cout << "setKick: kickSpeed: Power: " << kickDir.norm()/KICK_POWER_RATE << " Angle: " << rad_in_deg(kickDir.arg()) << endl;
    #endif

      sim_ball.setze_v(sim_ball.frage_v() + kickDir);

//      cout << "setKick: ballSpeed: x1: " << sim_ball.frage_v().x1 << " x2: " << sim_ball.frage_v().x2 << endl;

      sim_ich.simulieren(1);
      sim_ball.simulieren(1);
      sim_ich.sicher ();
      sim_ball.sicher ();

      if((sim_ball.frage_p() - sim_ich.frage_p()).norm() < PLAYER_RADIUS + BALL_RADIUS) {
//	sim_ich.setze_v(-0.1 * sim_ich.frage_v());
	sim_ball.setze_v(-0.1 * sim_ball.frage_v());
#ifdef DEBUG
	cout << "setKick: Ball and Player collided" << endl;
#endif
      }

//      cout << "setKick: AFTER:" << endl;
//      cout << "setKick: myPos    : x1: " << sim_ich.frage_p().x1 << " x2: " << sim_ich.frage_p().x2 << endl;
//      cout << "setKick: mySpeed  : x1: " << sim_ich.frage_v().x1 << " x2: " << sim_ich.frage_v().x2 << endl;
//      cout << "setKick: ballPos  : x1: " << sim_ball.frage_p().x1 << " x2: " << sim_ball.frage_p().x2 << endl;
//      cout << "setKick: ballSpeed: x1: " << sim_ball.frage_v().x1 << " x2: " << sim_ball.frage_v().x2 << endl;
		
      
  }

/**
 * Testet ob die angegebene kickDir beim simulieren zu keiner Kollision
 * fuehrt. Falls damit eine Kollision verursacht wird, wird das
 * Programm terminiert.
 */
  void AdvancedSkills::testAssertion(Geschwindigkeit kickDir) {
      return;
      Ball projectedBall;
    
      assert(!(kickDir == NULL_VEKTOR));

      // Fehlerfall testen (ob der Zwischenkick auch eine Kollision verursacht);
      projectedBall = sim_ball;
      projectedBall.setze_v(projectedBall.frage_v() + kickDir);
      if(collides(projectedBall, (Spieler)sim_ich) >= 0) {
	projectedBall = sim_ball;
	projectedBall.setze_v(projectedBall.frage_v() + kickDir);
	Punkt tempMyPos = sim_ich.frage_p() + sim_ich.frage_v();
	Punkt tempBallPos =  sim_ball.frage_p() + projectedBall.frage_v();
	cout << "testAssertion: Assertion failed: myPos    : x1: " << tempMyPos.x1 << " x2: " << tempMyPos.x2 << endl;
	cout << "testAssertion: Assertion failed: ballPos  : x1: " << tempBallPos.x1 << " x2: " << tempBallPos.x2 << endl;
	cout << "testAssertion: Assertion failed: DistanceBetweenPlayerAndBall: " << (tempMyPos - tempBallPos).norm() << endl;
	
	assert(0);    
      }   
  }

/**
 * Berechnet den Winkel um den sich, der eigene Spieler drehen muss,
 * um genau auf den angegebenen Punkt (absolut) zu schauen.
 */
  Winkel AdvancedSkills::ComputeTurnAngle(Punkt point, Punkt myPos = NULL_POINT) {
    Winkel turnAngle;
    Winkel viewAngle;

    sim_ich = myworld->frage_selbst();

    if(myPos == NULL_POINT) myPos = sim_ich.frage_p();

    turnAngle = rad_in_deg( (point - myPos).arg() ); 
    viewAngle = rad_in_deg( sim_ich.frage_richtung().arg() );
#ifdef DEBUG
    cout << "ComputeTurnAngle: DestinationVector: x: " << (point-myPos).x1 << " y: " << (point-myPos).x2 << endl;
    cout << "ComputeTurnAngle: angle(absolut): " << turnAngle;
    cout << "   eigene Blickrichtung :" << viewAngle;
#endif
    // eigene Blickrichtung abziehen -> rel. Winkel zum Ball;

    turnAngle -= viewAngle;
    if(turnAngle > 180) turnAngle -= 360;
    if(turnAngle < -180) turnAngle +=360;
#ifdef DEBUG
    cout << "   turnAngle: " << turnAngle << endl;
#endif
    return turnAngle;
  }

/**
 * Bestimmt die Vorzugsrichtung fuer den angegebenen Punkt.
 */

Vektor AdvancedSkills::ComputeBestWatchDirection (Punkt point) 
{
Vektor direction;
Punkt pos_self;
Punkt pos_goal;
char side;

  side = myworld->frage_seite ();
/*  if (myworld->frage_spielzustand () == KICK_OFF)
  {
    if (side == linke_seite)
      direction.x1 = -10.0;
    else
      direction.x1 = 10.0;
    direction.x2 = 0.0;
    return direction;
  } */

  // KICK_OFF Situation einbauen
  switch (ComputeRegion (point))
  {
    case REGION_I:
      direction.x1 = 0.0 - point.x1;
      direction.x2 = 34.0 - point.x2;
      break;
    case REGION_II:
      direction.x1 = 0.0 - point.x1;
      direction.x2 = -34.0 - point.x2;
      break;
    case REGION_III:
      if (side == linke_seite)
        direction.x1 = 10.0;
      else
        direction.x1 = -10.0;
      direction.x2 = 0.0;
      break;
    case REGION_VIII:
    case REGION_IV:
      pos_goal = myworld->frage_ihr_tor ().p;
      pos_goal.x2 += GOAL_WIDTH/2.0;
      direction = pos_goal - point;
      if (side == linke_seite)
        direction.drehen (-deg_in_rad (40.0));
      else
        direction.drehen (deg_in_rad (40.0));
      break;
    case REGION_X:
    case REGION_V:
      pos_goal = myworld->frage_ihr_tor ().p;
      pos_goal.x2 -= GOAL_WIDTH/2.0;
      direction = pos_goal - point;
      if (side == linke_seite)
        direction.drehen (deg_in_rad (40.0));
      else
        direction.drehen (-deg_in_rad (40.0));
      break;
    case REGION_VI:
      if (side == linke_seite)
        direction.x1 = 52.5 - point.x1;
      else
        direction.x1 = -52.5 - point.x1;
      direction.x2 = -point.x2;
      break;
    case REGION_VII:
      if (side == linke_seite)
        direction.x1 = 10.0;
      else
        direction.x1 = -10.0;
      direction.x2 = -10.0;
      break;
    case REGION_IX:
      if (side == linke_seite)
        direction.x1 = 10.0;
      else
        direction.x1 = -10.0;
      direction.x2 = 10.0;
      break;
  }
  return direction;
}

/**
 * Diese Routine fuehrt Kommandos aus um einen nah gelegenen Ball zu erreichen.
 * Wenn der Spieler noch die Moeglichkeit hat sich umzudrehen, dann dreht er
 * sich in Vorzugsrichtung.
 */
  void AdvancedSkills::FetchNearBall() {
      
      int firstDash = 0, secondDash = 0, thirdDash = 0;
      double turnAngle = 0.0;
      Punkt myPos, ballPos, myAcc;
      Winkel viewAngle;
      double distance;

      sim_ball = myworld->frage_ball();
      sim_ich = myworld->frage_selbst();
      
      sim_ball.simulieren(3);  // Ball bis auf die Pos. simulieren, wo ich ihn abfangen will;
      sim_ich.simulieren(3);   // mich simulieren, wenn ich nichts mache;
      
      myPos = sim_ich.frage_p();
      ballPos = sim_ball.frage_p();
      viewAngle = sim_ich.frage_richtung().arg();

      distance = (ballPos - myPos).norm();

#ifdef DEBUG
      cout << "FetchNearBall: Start: ballPos: x: " << ballPos.x1 << " y: " << ballPos.x2 << endl;
      cout << "FetchNearBall: Start: myPos: x: " << myPos.x1 << " y: " << myPos.x2 << endl;
#endif

      if(distance >= NEAR_BALL_RANGE) {
	// Spieler wird noch zu weit vom Ball weg sein -> hinlaufen;
#ifdef DEBUG
	cout << "FetchNearBall: Ich muss zum Ball laufen" << endl;
#endif

	firstDash = computeDash(myPos, ballPos, viewAngle, 3);

	// den ersten dash einrechnen;
	myAcc.setzen(viewAngle, firstDash * KICK_POWER_RATE);
	myPos += 1.56 * myAcc;

#ifdef DEBUG
	cout << "FetchNearBall: firstDash: power: " << firstDash << endl;
	cout << "FetchNearBall: after firstDash: myPos: x: " << myPos.x1 << " y: " << myPos.x2 << endl;
#endif


	if(firstDash == 100) {
	  // habe vollen dash gemacht -> zweiter dash wird benoetigt;
	  secondDash = computeDash(myPos, ballPos, viewAngle, 2);
	  
	  // den zweiten dash einrechnen;
	  myAcc.setzen(viewAngle, secondDash * KICK_POWER_RATE);
	  myPos += 1.4 * myAcc;

#ifdef DEBUG
	  cout << "FetchNearBall: secondDash: power: " << secondDash << endl;
	  cout << "FetchNearBall: after secondDash: myPos: x: " << myPos.x1 << " y: " << myPos.x2 << endl;
#endif
	}

	// ich habe eventuell keine Zeit mehr zum drehen, da ich eventuell;
	// vollen dash gemacht habe, aber erst mal sehen;	  
	distance = (ballPos - myPos).norm();
	if(distance <= NEAR_BALL_RANGE) {
	  // alles ok -> ich drehe mich in Vorzugsrichtung;
	  turnAngle = ComputeTurnAngle(myPos + ComputeBestWatchDirection(myPos), myPos);
#ifdef DEBUG
	  cout << "FetchNearBall: turn: angle: " << turnAngle << endl;
#endif
	}
	else if(distance <= KICK_RANGE) {
	  // hier steht die Frage: drehen oder nochmal beschleunigen;
	  switch(myworld->frage_rolle()) {
	  case ROLLE_TORWART:
	  case ROLLE_LIBERO:
	    // diese beiden sollen noch mal beschleunigen;
	    thirdDash = computeDash(myPos, ballPos, viewAngle, 1);
	    myAcc.setzen(viewAngle, secondDash * KICK_POWER_RATE);
	    myPos += 1.0 * myAcc;
#ifdef DEBUG
	    cout << "FetchNearBall: thirdDash: power: " << thirdDash << endl;
	    cout << "FetchNearBall: after thirdDash: myPos: x: " << myPos.x1 << " y: " << myPos.x2 << endl;
#endif
	    break;
	  default:
	    // alle anderen sollen sich in Vorzugsrichtung drehen;
	    turnAngle = ComputeTurnAngle(myPos + ComputeBestWatchDirection(myPos), myPos);
#ifdef DEBUG
	    cout << "FetchNearBall: turn: angle: " << turnAngle << endl;
#endif
	    break;
	  }
	}
	else {
	  // ich bin noch nicht in kick-Reichweite -> also nochmal beschleunigen;
	  thirdDash = computeDash(myPos, ballPos, viewAngle, 1);
	  myAcc.setzen(viewAngle, secondDash * KICK_POWER_RATE);
	  myPos += 1.0 * myAcc;
#ifdef DEBUG
	  cout << "FetchNearBall: thirdDash: power: " << thirdDash << endl;
	  cout << "FetchNearBall: after thirdDash: myPos: x: " << myPos.x1 << " y: " << myPos.x2 << endl;
#endif
	}
      }
	
      else {
	// Spieler wird zu nah am Ball sein -> weglaufen;
	;
#ifdef DEBUG
	cout << "FetchNearBall: Ich muss vom Ball weglaufen, da sonst Kollision" << endl;
#endif

	turnAngle = ComputeTurnAngle(myPos + ComputeBestWatchDirection(myPos), myPos);

#ifdef DEBUG
	cout << "FetchNearBall: turn: angle: " << turnAngle << endl;
#endif
      }
      
      // jetzt habe ich alle Kommandos ausgerechnet, jetzt kann ich sie also absenden;
      if(thirdDash == 0) SimpleDashTurn(firstDash, secondDash, turnAngle, IGNORE);
      else {
	SimpleDash(firstDash, IGNORE);
	SimpleDash(secondDash, IGNORE);
	SimpleDash(thirdDash, IGNORE);
      }

      return;
  }

/**
 * Berechnet den optimalen dash den der Spieler ausfuehren muss um den Abstand 
 * zur simulierten Ballposition zu minimieren.
 */
  int AdvancedSkills::computeDash(Punkt myPos, Punkt ballPos, Winkel viewAngle, int simulateTicks) {
      double distanceNew, distanceOld;
      int dash;
      Punkt mySimPos, myAcc;
      
      distanceNew = (ballPos - myPos).norm();
      distanceOld = distanceNew + 0.1;
      dash = 0;
      while( distanceNew > NEAR_BALL_RANGE && distanceNew < distanceOld && dash < 100) {
	dash += 10;
	mySimPos = myPos;
	myAcc.setzen(viewAngle, (double)dash * KICK_POWER_RATE);

	switch(simulateTicks) {
	case 3:
	  mySimPos += 1.56 * myAcc; // 1.56=1+0.4+0.16;
	  break;
	case 2:
	  mySimPos += 1.4 * myAcc; // 1.4=1+0.4;
	  break;
	case 1:
	  mySimPos += 1.0 * myAcc; // 1.0=1;
	  break;
	default:	
	  mySimPos += 1.0 * myAcc; // 1.0=1;
	  break;
	}
	
	distanceOld = distanceNew;
	distanceNew = (ballPos - mySimPos).norm();
	}
      return dash;
  }

// Jan

int AdvancedSkills::ComputeRegion (Punkt pos_self)
{
  if (myworld->frage_seite () == rechte_seite)
    pos_self.x1 = -pos_self.x1;
  if (pos_self.x1 <= 15)
  {
    if (pos_self.x2 >= AREA_O_1)
      return REGION_VII;
      
    if (pos_self.x2 <= AREA_U_1)
      return REGION_IX;
      
    if (pos_self.x1 <= -20 && pos_self.x2 <= 20 && pos_self.x2 >= 5)
      return REGION_I;

    if (pos_self.x1 <= -20 && pos_self.x2 <= -5 && pos_self.x2 >= -20)
      return REGION_II;

    return REGION_III;
  }
  else
  {
    if (pos_self.x2 >= AREA_O_1)
      return REGION_VIII;
     
    if (pos_self.x2 <= AREA_U_1)
      return REGION_X;
  
  double m = (GOAL_WIDTH/2.0 - 34.0)/34.0;
  double n = 34.0 - 18.5 * m;
    if (pos_self.x2 > m * pos_self.x1 + n)
      return REGION_IV;

    m = (-GOAL_WIDTH/2.0 + 34.0)/34.0;
    n = -34.0 - 18.5 * m;
    if (pos_self.x2 < m * pos_self.x1 + n)
      return REGION_V;

    return REGION_VI;
  }	
}  


////////////////////////////////////////////
// Erweiterte Faehigkeiten: Sprechen      //
////////////////////////////////////////////

/*void AdvancedSkills::SayStamina (int stat=IGNORE)
{
char message [50];
Punkt pos_self;

  pos_self = myworld->frage_selbst ().frage_p ();
  sprintf (message, "STAMINA_%i_%i_%f_%f",
                    myworld->frage_selbst().frage_nummer(),
                    (int) myworld->frage_selbst().frage_stamina(),
                    (float) pos_self.x1,
                    (float) pos_self.x2
                    );
  mysender->SetCommand (SAY, message, stat);
}*/

void AdvancedSkills::CollectInfos (int stat=IGNORE)
{
  //mysender->SetCommand (TURN, 90.0, stat);
  SimpleTurn (-88.0, stat);
}

void AdvancedSkills::ShareInfos (int stat=IGNORE)
{
char message [512];
char code [20];
Punkt pos;
Spieler player;
int i;

  pos = myworld->frage_selbst ().frage_p ();
  sprintf (message, "%i_%i_%i_%i", 
           myworld->frage_selbst ().frage_nummer ()+1,
           (int) myworld->frage_selbst ().frage_stamina (),
           (int) rint (pos.x1),
           (int) rint (pos.x2)
           );
  for (i=0; i<=9; i++)
  {
    player = myworld->frage_spieler (i, team_wir);
    if (player.frage_wert () <= 1)
    {
      pos = player.frage_p ();
      sprintf (code, "_o_%i_%i_%i", player.frage_nummer ()+1, (int) rint (pos.x1), (int) rint (pos.x2));
      strcat (message, code);
    }
  }          
  for (i=0; i<=10; i++)
  {
    player = myworld->frage_spieler (i, team_sie);
    if (player.frage_wert () <= 1)
    {
      pos = player.frage_p ();
      sprintf (code, "_t_%i_%i_%i", player.frage_nummer ()+1, (int) rint (pos.x1), (int) rint (pos.x2));
      strcat (message, code);
    }
  }
  #ifdef DEBUG          
  cout << message << endl;
  #endif
  mysender->SetCommand (SAY, message, stat);
}
