/////////////////////////////////////////////////////////////////////////////////////////
/////  sensors.h - Klassen zur Kapselung der sensorischen Daten

#ifndef _SENSOR_H_
#define _SENSOR_H_

#if __GNUG__ >= 2
#  pragma interface
#endif

#include "comm.h"
#include "alinkli.h"

#define FLAG_LEFTBOTTOM 0
#define FLAG_LEFTTOP 1
#define FLAG_CENTERBOTTOM 2
#define FLAG_CENTERTOP 3
#define FLAG_RIGHTBOTTOM 4
#define FLAG_RIGHTTOP 5
#define FLAG_PENALTY_LEFTBOTTOM 6
#define FLAG_PENALTY_LEFTCENTER 7
#define FLAG_PENALTY_LEFTTOP 8
#define FLAG_PENALTY_RIGHTBOTTOM 9
#define FLAG_PENALTY_RIGHTCENTER 10
#define FLAG_PENALTY_RIGHTTOP 11
#define FLAG_ANONYMOUS 12
#define FLAG_NOTVALID 13

#define LINE_LEFT 0
#define LINE_CENTER 1
#define LINE_RIGHT 2
#define LINE_TOP 3
#define LINE_BOTTOM 4
#define LINE_ANONYMOUS 5
#define LINE_NOTVALID 6

#define PLAYER_NUMBER_NOTVALID -1
#define PLAYER_TEAM_NOTVALID "team_notvalid"
#define PLAYER_OURPLAYER_NOTVALID -1

#define MESSAGETYPE_SEE 1
#define MESSAGETYPE_HEAR 2
#define MESSAGETYPE_ERROR 3

#define MESSAGESENDER_REFEREE 888.0
#define MESSAGESENDER_SELF 999.0
#define MESSAGESENDER_NOTVALID -999.0

#define DIST_NOTVALID -999.0
#define ANGLE_NOTVALID -999.0
#define DISTCHANGE_NOTVALID -999.0
#define DIRCHANGE_NOTVALID -999.0

class RelPos
{
public:
  RelPos ();
  RelPos & operator=(RelPos &x); // Zuweisungsoperator;
  
  void SetDistance (float dist);
  void SetAngle (float angle);
  void SetDistChange (float distchange);
  void SetDirChange (float dirchange);
  void SetAll (float dist, float angle, float distchange, float dirchange);
  
  float GetDistance ();
  float GetAngle ();
  float GetDistChange ();
  float GetDirChange ();
  void GetAll (float* dist, float* angle, float* distchange, float* dirchange);
	
  int IsValid ();
  void SetValid (int valid);
  
protected:
  float m_fDistance, m_fAngle, m_fDistChange, m_fDirChange;
  int m_bValid;
};

class Player:public RelPos
{
public:
  Player ();
  Player (char* team, int number);
  ~Player ();
  //Player & operator=(Player &x); // Zuweisungsoperator;
  
  void SetPlayerNumber (int number);
  void SetTeamName (char* team);
  
  int GetPlayerNumber ();
  char* GetTeamName ();
  
private:
  int m_nNumber;
  char m_sTeam [MAXMESG];
};	

class Sensors 
{
public:
  Sensors ();
  Sensors (Comm* communication);
  ~Sensors ();
  
  void InitSensorsComm (Comm* communication);
  int ListenAndWatch ();
  int ListenAndWatch (int usec);
  int ParseSensoryData (char* buf);
  
  RelPos GetBall ();
  RelPos GetRightGoal ();
  RelPos GetLeftGoal ();
  RelPos GetAnonymousGoal ();
  RelPos GetAFlag (int flagid);
  RelPos GetALine (int lineid);
  ALinkList <Player> *GetPlayers ();
  
  int GetMessageType ();
  int GetTime ();
  float GetMessageSender ();
  char* GetMessage ();
  
private:
  void ParseSights (char* message);
  void ParseSounds (char* message);
  void SeeMarker (char* marker, float dist, float angle, float distchange, float dirchange);
  void SeeBall (float dist, float angle, float distchange, float dirchange);
  void SeePlayer (float dist, float angle, float distchange, float dirchange, char* team, int number);
  
  RelPos m_cBall;
  RelPos m_cRightGoal;
  RelPos m_cLeftGoal;
  RelPos m_cAnonymousGoal;
  RelPos m_cFlags [FLAG_NOTVALID+1];
  RelPos m_cLines [LINE_NOTVALID+1];
  ALinkList <Player> m_cPlayers;
  
  Comm* m_pcComm;
  int m_bOwnComm;
  
  int m_nMessageType;
  char m_sMessage [MAXMESG];
  char m_sError [MAXMESG];
  float m_fSender;
  int m_nTime;
};

#endif
