#ifndef _WELTMODELL
#define _WELTMODELL

//#define DEBUG


#include<limits.h>
#include"commando.h"
#include"spielfe.h"
#include "baseskills.h"
#include "Configuration.h"

typedef Vektor Geschwindigkeit;



// Klasse Unzuverlaessigkeit

const long wert_sicher = 0;
const long wert_unsicher = 6;

class Unzuverlaessigkeit
{
private:
  long wert;
  
public:
  inline Unzuverlaessigkeit(void) // Am Anfang auf 'unsicher' gesetzt
  {unsicher();}


  inline void erhoehen(void) // Erhoeht die Unzuverlassigkeit um 1 Stufe
  {if(wert < wert_unsicher) wert++;}
  
  inline void sicher(void) // Setzt den Wert auf 'sicher'
  {wert = wert_sicher;}
  
  inline void unsicher(void) // Setzt den Wert auf 'unsicher'
  {wert = wert_unsicher;}
  
  inline long frage_wert(void) // liefert den Wert
  {return wert;}
  
  inline int ist_unsicher(void) const // Ob die Angabe unsicher ist
  {return wert == wert_unsicher;}

  inline int ist_sicher(void) const // Ob die Angabe sicher ist
  {return wert == wert_sicher;}
};



// Klasse Uhr

typedef long Zeit;

class Uhr
{
private:
  Zeit	      zeit;
  static Zeit jetzt;
  
public:
  inline Uhr(Zeit init = 0) : zeit(init) {}  

  inline Zeit frage_zeit(void) const {return zeit;} // Was unsere Uhr anzeigt
  static Zeit frage_jetzt(void); // Die globale Jetzt-Zeit zur Zeit des Aufrufs
  static void erhoehe_jetzt(void); // Die globale Jetzt-Zeit weiterzahlen
  static void setze_jetzt(Zeit neu); // Die globale Jetzt-Zeit neu setzen
  inline void setze_zeit(Zeit neu) {zeit = neu;}
  void aktualisiere(void); // Auf Jetzt setzen
};



// Klasse Position

class Position
{
private:
  Punkt p;
  
public:
  Position(void) {p.x1 = p.x2 = 0;}

  inline void setze_p(const Punkt &neu) {p = neu;}
  inline const Punkt &frage_p(void) const {return p;}
};



// Klasse Bewegter_Koerper

class Bewegter_Koerper : public Unzuverlaessigkeit,public Position
{
private:
  Uhr		  uhr;
  Koerper	  verlangsamung; // Rate: 1 bedeutet keine Verlangsamung
  Koerper	  power_rate; // s. N. Itsukis dash_power_rate bzw. kick_power_rate
  Koerper         max_speed;  // max. Geschwindigkeit
  Geschwindigkeit v;
  
public:
  // friend Koerper metrik(const Bewegter_Koerper &x,const Bewegter_Koerper &y);
  
  Bewegter_Koerper(Koerper decay,Koerper rate,Koerper speed)
  {v.x1 = v.x2 = 0; verlangsamung = decay; power_rate = rate; max_speed=speed;}

  
  virtual void beschleunigen(Koerper power,const Vektor &richtung);
  // Beschleunigung in Richtung des jetzigen Geschwindigkeitsvektors in angegebener Richtung

  virtual void drehen(Winkel teta);
  // Drehen um den Winkel teta, negative Winkel: Drehung nach links, positive: nach rechts
  
  virtual void setze_v(const Geschwindigkeit &neu);
  
  inline void initialisiere_uhr(Zeit zeit) // Setzen der Uhr
  {uhr.setze_zeit(zeit);}
  
  inline const Geschwindigkeit &frage_v(void) const {return v;}

  inline Zeit frage_zeit(void) const {return uhr.frage_zeit();}
  inline Zeit frage_jetzt(void) const {return uhr.frage_jetzt();}
	
  void simulieren(int anzahl_schritte);
  // Bewegung des Objektes fur entspr. Anzahl Schritte vorausberechnen.
  // Dabei wird auch die Geschwindigkeit (Reibung etc.) verandert.

  void simulieren_bis(Zeit bis);
  // Simulieren bis zum angegebenen Zeitpunkt
};

// Koerper metrik(const Bewegter_Koerper &x,const Bewegter_Koerper &y);
// Abstandsmessung zur Einordnung neuer Sensorinformationen



// Klasse Spielerkennzeichnung

const int team_wir = 1;
const int team_sie = 2;
const int team_unsicher = -1;
const int nummer_unsicher = -1;
// Ubrige Nummern von 0 bis 10

class Spielerkennzeichnung
{
private:
  int nummer;
  int team;
  
public:
  Spielerkennzeichnung(int init_nummer = nummer_unsicher,int init_team = team_unsicher)
  {nummer = init_nummer,team = init_team;}

  
  // Funktionen werden je nur einmal gerufen (zur Initialisierung);
  inline void initialisiere_nummer(int init_nummer)
  {nummer = init_nummer;}

  inline void initialisiere_team(int init_team)
  {team = init_team;}

  inline int frage_nummer(void) const {return nummer;}
  // Spielernummern laufen von 0 bis 10

  inline int frage_team(void) const {return team;}
};



const int sk_ungleich = 0;	// Die Kennzeichnung ist verschieden (-> die Spieler auch)
const int sk_gleich = 1;	// Die Kennzeichnung stimmt uberein (-> die Spieler sind gleich)
const int sk_nicht_vergleichbar = 2; // Wir konnen keine Aussage machen (-> beide Falle konnen eintreten)

int vergleiche(const Spielerkennzeichnung &sk1,const Spielerkennzeichnung &sk2);
// vergleicht zwei Spielerkennzeichnungen und liefert einen der o.g. Werte



// Klasse Spieler

//const Koerper player_decay = 0.4;
//const Koerper dash_power_rate = 0.01;

class Spieler : public Bewegter_Koerper, public Spielerkennzeichnung
{
private:
  float stamina;  // Stamina des Spielers
public:
  Spieler(int init_nummer = nummer_unsicher,int init_team = team_unsicher)
    : Bewegter_Koerper(globalConf->playerDecay,globalConf->dashPowerRate,globalConf->playerSpeedMax),Spielerkennzeichnung(init_nummer,init_team),stamina(globalConf->playerStaminaMax)
  {}

  void update_stamina(float wert); // Wert wird zu Stamina addiert
  void setze_stamina(float wert);  // Stamina auf Wert setzen
  float frage_stamina() const; // Stamina abfragen  
};



// Klasse Spieler_Selbst

class Spieler_Selbst : public Spieler
{
private:
  Vektor richtung;	// Einheitsvektor, der die Blickrichtung angibt
  
public:
  Spieler_Selbst(void); // Weltmodell mus konstruiert sein.

  
  inline const Vektor &frage_richtung(void) const {return richtung;}
	
  void beschleunigen(Koerper power);
  // Beschleunigung in Richtung des jetzigen Richtungsvektors

  void drehen(Winkel teta);
  // Drehen um den Winkel teta, negative Winkel: Drehung nach links, positive: nach rechts

  Vektor frage_richtung();
  // Richtung abfragen
  
  void setze_richtung(const Vektor &neu);
  // neue Richtung einstellen und dabei die Geschwindigkeit mit in die neue Richtung umsetzen

};



#include"sensori.h"

// Abstandsmessung zur Einordnung neuer Sensorinformationen
const Koerper metrik_unendlich = Koerper_Max;

Koerper metrik(Sensor_Spieler &x,Spieler &y);



// Klasse Ball
// Werte aus globalConf werden benutzt
//const Koerper ball_decay = 0.96;
//const Koerper kick_power_rate = 0.013;

class Ball : public Bewegter_Koerper
{
public:
  inline Ball(void) : Bewegter_Koerper(globalConf->ballDecay,globalConf->kickPowerRate,globalConf->ballSpeedMax) {}
};



// Klasse Umgebung

class Umgebung
{
private:
  void aktualisiere_spieler(Sensorinformation &s);
  void aktualisiere_selbst(Sensorinformation &s);
  void aktualisiere_ball(Sensorinformation &s);
  
public:
  Spieler_Selbst selbst;
  Spieler        freunde[10];
  Spieler	 gegner[11];
  Ball		 ball;
  double   theOffsideLineX;
	
  Umgebung(void); // Weltmodell mus konstruiert sein.

  void aktualisiere(Sensorinformation &s);
  void simulieren_bis(Zeit bis);
  // Gesamtsituation vorausrechnen.
  
  double offsideLineX();
  double computeOffsideLineX();
};




// Klasse Weltmodell

const int weltmodell_tiefe = 5;

#define KEINE_INITIATIVE -1
#define INITIATIVE_SIE    0    
#define INITIATIVE_WIR    1

#define NICHTS_NEUES      0
#define INFOS_GESEHEN     1
#define INFOS_GEHOERT     2
#define SEQUENZ_BEENDET   3
#define SONSTIGE_GRUENDE  4

// Konstanten: s. spielfeld.h

class Weltmodell
{
private:
  int	   seite, nummer;
  char	   name[256];
  Zeit   m_nLetzteSensorZeit;

  int rolle, stimmung, spielzustand, entscheidung, aufstellung, initiative, sichtbreite, sichtqualitaet;
  double params [6];

  Punkt stammpos_spiel, stammpos_anstoss, stammpos_ecke, stammpos_abstoss;

  Umgebung *situation[weltmodell_tiefe];
  Umgebung *neue_umgebung(void);
  
public:
  int blockierenderGegnerIndex;
  Weltmodell(void);  
  ~Weltmodell(void);
  
  void setze_letzte_sensor_zeit (Zeit zeit);
  Zeit frage_letzte_sensor_zeit ();
  
  void initialisiere_seite(char seite); // eine der Konstanten von oben
  void initialisiere_teamname(const char *name); // name wird kopiert.
  void initialisiere_nummer(int nummer); // Eigene Nummer setzen. Parameter ist eine Nummer von 1 bis 11

  void setze_rolle(int neue_rolle); // Rolle setzen
  int  frage_rolle(void); // Rolle abfragen

  void setze_stimmung(int neue_stimmung); // Stimmung setzen
  int  frage_stimmung(void); // Stimmung abfragen

  void setze_initiative(int wer); // Initiative setzen - Werte siehe oben
  int frage_initiative(void); // Initiative abfragen

  void setze_stammpos_anstoss(Punkt neue_stammpos_anstoss); // Anstossposition setzen
  Punkt frage_stammpos_anstoss(void); // Anstossposition abfragen

  void setze_stammpos_spiel(Punkt neue_stammpos_spiel); // Spiel-Stammposition setzen
  Punkt frage_stammpos_spiel(void); // Spiel-Stammposition abfragen

  void setze_stammpos_ecke (Punkt neue_stammpos_ecke);
  Punkt frage_stammpos_ecke ();

  void setze_stammpos_abstoss (Punkt neue_stammpos_abstoss);
  Punkt frage_stammpos_abstoss ();
  
  void setze_aufstellung (int neue_aufstellung);
  int frage_aufstellung ();

  int neue_entscheidung(void); // neue Entscheidung treffen ? -- Grund fuer neue Entscheidung wird zurueckgegeben
  void entscheidung_getroffen(void); // Entscheidung wurde getroffen
  void neu_entscheiden(int grund); // neue Entscheidung wird notwendig

  void setze_spielzustand(int neuer_spielzustand); // Spielzustand setzen - vgl. Nachrichten vom Schiedsrichter
  int frage_spielzustand(void); // Spielzustand abfragen
  
  inline const char *frage_teamname(void) const {return name;}
  // Gelieferter Zeiger verweist auf ein lokales Objekt und darf nicht geloscht werden.
  // Zeiger ist stets gesetzt, also != 0.

  inline char frage_seite(void) const {return seite;}
  // liefert eine der o.g. Konstanten

  inline int frage_nummer(void) const {return nummer;}
  // liefert die eigene Nummer, ein Wert von 0 bis 10.
  
  void neue_sensorinformation(Sensors *s);

  void extrapoliere(Zeit bis); // bis zur angegebenen Zeit extrapolieren
  
  inline Umgebung *frage_aktuelle_situation(void)
    // liefert die neuesten im Weltmodell vorliegenden Informationen
    // u.U. wird 0 geliefert, wenn keine Information vorliegt.
  {return situation[0];}

  inline Umgebung *frage_fruehere_situation(int nummer)
    // liefert eine fruhere Situation. nummer ist im Bereich von 0 bis weltmodell_tiefe - 1.
    // 0 ist die aktuelle, weltmodell_tiefe - 1 die alteste Information. U.U. wird 0
    // geliefert, wenn die Information nicht soweit zuruckreicht
  {
    if(nummer < 0 || nummer >= weltmodell_tiefe)
      exception("Weltmodell::frage_fruehere_situation: Illegaler Parameter");
    return situation[nummer];
  }
  
  double frage_param (int parnumber)
  {
    return params [parnumber];
  }
  
  void setze_param (int parnumber, double param)
  {
    params [parnumber] = param;
  }
  
  int frage_sicht_breite ()
  {
    return sichtbreite;
  }
  
  void setze_sicht_breite (int sichtbr)
  {
    sichtbreite = sichtbr;
  }
  
  int frage_sicht_qualitaet ()
  {
    return sichtqualitaet;
  }
  
  void setze_sicht_qualitaet (int sichtqual)
  {
    sichtqualitaet = sichtqual;
  }
  
  Spieler &frage_spieler(int spielernummer,int team,int zeitstufe = 0);
  // spielernummer ist eine Nummer von 0 bis 10, falls nach einem gegnerischen
  // Spieler gefragt wird, sonst eine Nummer von 0 bis 9.
  // team ist team_wir oder team_sie, Konstanten siehe oben unter Spielerkennzeichnung
  // zeitstufe ist ein Wert von 0 bis weltmodell_tiefe - 1, der default-Parameter
  // bezeichnet die jungste Zeitstufe, d.h. die aktuellste gehaltene Information.
  // Die gelieferte Referenzvariable darf nicht zwischengespeichert werden.

  Ball &frage_ball(int zeitstufe = 0);
  // zeitstufe s. oben.
  // Die gelieferte Referenzvariable darf nicht zwischengespeichert werden.

  Spieler_Selbst &frage_selbst(int zeitstufe = 0);
  // zeitstufe s. oben.
  // Die gelieferte Referenzvariable darf nicht zwischengespeichert werden.
  
  int berechne_relativ(Relative_Koordinate &delta,const Punkt &p,int zeitstufe = 0);
  // Berechnen der relativen Angaben bzgl. des Spielers zur gegebenen
  // Zeit zum Punkt p. Ergebnis wird in delta abgelegt.
  // Ergebniswert wie berechne_relative_koordinate in Klasse Spielfeld.

  int berechne_relativ(Relative_Koordinate &delta,const Spieler_Selbst ich, const Punkt &p,int zeitstufe = 0);
  // Berechnen der relativen Angaben bzgl. des angegebenen Spielers
  // Zeit zum Punkt p. Ergebnis wird in delta abgelegt.
  // Ergebniswert wie berechne_relative_koordinate in Klasse Spielfeld.   
    
  const Tor &frage_ihr_tor(void) const;

  const Tor &frage_unser_tor(void) const;
    
  friend void update_by_command(cmd_struct cmd);
};

// Funktionen zum Aufruf bei entspr. Ereignis

#endif
